#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <mpi.h>
#include <omp.h>
#include <signal.h>
#include <unistd.h>

// Skeleton for the problem of finding the maximum distance of equal numbers in a matrix
// Spanish Parallel Programming Contest 2012

void initialize(int *m,int t,int maximo)
{
  int i;

// Integer numbers are generated randomly
  for(i=0;i<t*t;i++)
  {  
    m[i] =(int) (((1.*rand())/RAND_MAX)*maximo);
  }
}

void escribir(int *m, int t){
  int i, j;

  for (i = 0; i < t; i++)
  {
    for (j = 0; j < t; j++)
    {  
      printf("%d ",m[i*t+j]);
    }
    printf("\n");
  }
  printf("\n");
}

void escribirresult(int *m, int t){
  int i;

  for (i = 0; i < t; i++)
  {
    printf("%d ",m[i]);
  }
  printf("\n");
}

/*
c
c     mseconds - returns elapsed milliseconds since Jan 1st, 1970.
c
*/
long long mseconds(void) {
  struct timeval t;
  gettimeofday(&t, NULL);
  return t.tv_sec*1000 + t.tv_usec/1000;
}

static void alarm_handler(int sig) {
  fprintf(stderr, "Time Limit Exceeded\n");
  abort();
}

void distancia(int t,int *a,int n,int *numeros,int *distancias,int nodo,int np);

int main(int argc,char *argv[]) {
  int nodo,np,i,j,N,cuantos,n,semilla,maximo;
  long long ti,tf,tt=0;
  int *a,*numeros,*distancias;

  FILE *stats_file = fopen("stats", "w");

  struct sigaction sact;
  sigemptyset(&sact.sa_mask);
  sact.sa_flags = 0;
  sact.sa_handler = alarm_handler;
  sigaction(SIGALRM, &sact, NULL);
  alarm(20);  /* time limit */
 
  MPI_Init(&argc,&argv);
  MPI_Comm_size(MPI_COMM_WORLD,&np);
  MPI_Comm_rank(MPI_COMM_WORLD,&nodo);

// The number of test cases is read
// and sent to all the processes
  if(nodo==0)
  {
    scanf("%d",&cuantos);
    MPI_Bcast(&cuantos,1,MPI_INT,0,MPI_COMM_WORLD);
  }
  else
  {
    MPI_Bcast(&cuantos,1,MPI_INT,0,MPI_COMM_WORLD);
  }
  for(i=0;i<cuantos;i++)
  {
    if(nodo==0)
    {
      scanf("%d",&N);                   // The first argument is the matrix size
      scanf("%d",&semilla);             // the second is the seed for random generation
      scanf("%d",&maximo);		// the maximum value of the ingeter; which will be generated between 0 and maximum-1
      scanf("%d",&n);              	// and the number of integers

// Space for the matrix, the nubmers and the distances
      a = (int *) malloc(sizeof(int)*N*N);
      numeros = (int *) malloc(sizeof(int)*n);
      distancias=(int *) malloc(sizeof(int)*n);
// Read the numbers
      for(j=0;j<n;j++)
        scanf("%d",&numeros[j]);

      srand(semilla);
      initialize(a,N,maximo);

#ifdef DEBUG
// While debugging the matrix and the numbers are written
      escribir(a,N);
      escribirresult(numeros,n);
#endif
    }
    MPI_Barrier(MPI_COMM_WORLD);
    ti=mseconds(); 
    distancia(N,a,n,numeros,distancias,nodo,np);
    MPI_Barrier(MPI_COMM_WORLD);
    tf=mseconds(); 
    if(nodo==0)
    {
// The time of the first input is not considered
      if(i!=0) tt+=tf-ti;
#ifdef DEBUG
// While debugging the time of each execution is written
      fprintf(stats_file, "%Ld\n", tf-ti);
#endif
// The results of each problem are written,
// even for the first problem
      escribirresult(distancias,n);
      free(a);
      free(numeros);
      free(distancias);
    } 
  }
  
  if(nodo==0)
  { 
    fprintf(stats_file, "%Ld\n", tt);
    fclose(stats_file);
  }
  MPI_Finalize();
  return 0;
}
